from flask import render_template, jsonify, abort
from flask.views import MethodView


class APIview(MethodView):
    def __init__(self, *args, **kwargs):
        view_args = kwargs.pop('view_args', {})
        self.config = view_args.get('config')
        super().__init__(*args, **kwargs)

    def get(self):
        assert self.config.ui in self.config._support_ui
        ui_file = f'{self.config.ui}.html'
        return render_template(ui_file, spec_url=self.config.filename)


class JSONview(MethodView):
    def __init__(self, *args, **kwargs):
        view_args = kwargs.pop('view_args', {})
        self.config = view_args.get('config')
        super().__init__(*args, **kwargs)

    def get(self, filename):
        if filename == self.config.filename:
            return jsonify(self.config.data)
        abort(404)



def get_converter_type_any(*args, **kwargs):
    """
    Handle converter type "any"

    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'array',
        'items': {
            'type': 'string',
            'enum': args,
        }
    }
    return schema


def get_converter_type_int(*args, **kwargs):
    """
    Handle converter type "int"

    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'integer',
        'format': 'int32',
    }
    if 'max' in kwargs:
        schema['maximum'] = kwargs['max']
    if 'min' in kwargs:
        schema['minimum'] = kwargs['min']
    return schema


def get_converter_type_float(*args, **kwargs):
    """
    Handle converter type "float"

    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'number',
        'format': 'float',
    }
    return schema


def get_converter_type_uuid(*args, **kwargs):
    """
    Handle converter type "uuid"

    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'string',
        'format': 'uuid',
    }
    return schema


def get_converter_type_path(*args, **kwargs):
    """
    Handle converter type "path"

    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'string',
        'format': 'path',
    }
    return schema


def get_converter_type_string(*args, **kwargs):
    """
    Handle converter type "string"

    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {
        'type': 'string',
    }
    for prop in ['length', 'maxLength', 'minLength']:
        if prop in kwargs:
            schema[prop] = kwargs[prop]
    return schema


def get_converter_type_default(*args, **kwargs):
    """
    Handle converter type "default"

    :param args:
    :param kwargs:
    :return: return schema dict
    """
    schema = {'type': 'string'}
    return schema



CONVERTER_TYPE_ANY = 'any'
CONVERTER_TYPE_INT = 'int'
CONVERTER_TYPE_FLOAT = 'float'
CONVERTER_TYPE_UUID = 'uuid'
CONVERTER_TYPE_PATH = 'path'
CONVERTER_TYPE_STRING = 'string'
CONVERTER_TYPE_DEFAULT = 'default'



from flask import abort, make_response, jsonify
from werkzeug.routing import parse_rule, parse_converter_args
from werkzeug.exceptions import default_exceptions

from flaskerk.const import CONVERTER_TYPE_ANY, CONVERTER_TYPE_INT, CONVERTER_TYPE_FLOAT, \
    CONVERTER_TYPE_UUID, CONVERTER_TYPE_PATH, CONVERTER_TYPE_STRING, CONVERTER_TYPE_DEFAULT
from flaskerk.converters import get_converter_type_any, get_converter_type_int, \
    get_converter_type_float, get_converter_type_uuid, get_converter_type_path, \
    get_converter_type_string, get_converter_type_default

CONVERTER_MAPPING = {
    CONVERTER_TYPE_ANY: get_converter_type_any,
    CONVERTER_TYPE_INT: get_converter_type_int,
    CONVERTER_TYPE_FLOAT: get_converter_type_float,
    CONVERTER_TYPE_UUID: get_converter_type_uuid,
    CONVERTER_TYPE_PATH: get_converter_type_path,
    CONVERTER_TYPE_STRING: get_converter_type_string,
    CONVERTER_TYPE_DEFAULT: get_converter_type_default
}


def abort_json(code: int, msg: str = ''):
    """
    abort as a JSON response

    :param int code: HTTP status code
    :param str msg: description for this abort
    """
    if not msg:
        assert code in default_exceptions
        msg = default_exceptions[code].description
    abort(make_response(jsonify(message=msg), code))


def get_converter(converter: str, *args, **kwargs):
    """
    Get conveter method from converter map

    :param converter: str: converter type
    :param args:
    :param kwargs:
    :return: return schema dict
    """
    return CONVERTER_MAPPING[converter](*args, **kwargs)


def parse_url(path: str):
    """
    Parsing Flask route url to get the normal url path and parameter type.

    Based on Werkzeug_ builtin converters.

    .. _werkzeug: https://werkzeug.palletsprojects.com/en/0.15.x/routing/#builtin-converters
    """
    subs = []
    parameters = []

    for converter, arguments, variable in parse_rule(path):
        if converter is None:
            subs.append(variable)
            continue
        subs.append(f'{{{variable}}}')

        args, kwargs = [], {}

        if arguments:
            args, kwargs = parse_converter_args(arguments)

        schema = get_converter(converter, *args, **kwargs)

        parameters.append({
            'name': variable,
            'in': 'path',
            'required': True,
            'schema': schema,
        })

    return ''.join(subs), parameters



from pydantic import BaseModel, Schema, validator
from werkzeug.exceptions import _aborter, default_exceptions
from werkzeug.exceptions import HTTPException as WerkzeugException
from werkzeug.http import HTTP_STATUS_CODES

from flaskerk.utils import abort_json


class HTTPValidationError(BaseModel):
    code: int = Schema(
        422,
        description='HTTP response status code',
    )
    description: str = 'Unprocessable Entity'


class HTTPException(BaseModel):
    """
    HTTP Exceptions.

    :param int code: HTTP status code
    :param str msg: description

    .. _code: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status

    If you are using default code_ like 403, 404, 500 ..., you can ignore the
    ``msg`` parameter and this module will use default description.

    If you want to define your own code and description, you must offer ``msg``.

    examples:

    .. code-block:: python

       from flaskerk import HTTPException
       code404 = HTTPException(code=404)
       # with customized discription
       code403 = HTTPException(code=403, msg='IP is blocked.')
       # customized code
       code777 = HTTPException(code=777, msg='bad luck')

       # abort
       code777.abort()
    """
    code: int = Schema(
        ...,
        gt=100,
        lt=1000,
        description='HTTP response status code',
    )
    msg: str = None

    @validator('msg', pre=True, always=True)
    def set_as_default(cls, v, values, **kwargs):
        if v is None:
            if 'code' not in values or values['code'] not in default_exceptions:
                raise ValueError('Invalid default HTTP status code.')
            v = default_exceptions[values['code']].description
        return v

    def __post_init_post_parse__(self):
        if self.code not in _aborter.mapping:
            _aborter.mapping[self.code] = type(
                f'NewHTTPException_{self.code}',
                (WerkzeugException, ),
                {'code': self.code, 'description': self.msg}
            )
            HTTP_STATUS_CODES[self.code] = self.msg

    def abort(self, msg=None):
        """
        abort as a JSON response
        """
        abort_json(self.code, msg or self.msg)



from functools import wraps
from pydantic import ValidationError, BaseModel
from flask import Blueprint, abort, request, jsonify, make_response, Flask

from flaskerk.config import Config
from flaskerk.view import APIview
from flaskerk.utils import abort_json, parse_url
from flaskerk.exception import HTTPException


class Flaskerk:
    """
    :param app: Flask app instance, you can register it later
    :param configs: key-value pairs in :class:`flaskerk.config.Config`

    example:

    .. code-block:: python

       from flask import Flask
       from flaskerk import Flaskerk

       app = Flask(__name__)
       api = Flaskerk(app, version='0.2', title='Machine Translation service')
    """

    def __init__(self, app=None, **configs):
        self.models = {}
        self.config = Config()
        self.config._spec = None
        for key, value in configs.items():
            setattr(self.config, key, value)

        if app:
            self._init_app(app)

    def _init_app(self, app):
        assert isinstance(app, Flask)
        self.app = app
        self.update_config(**self.app.config.get('OPENAPI', {}))
        self._register_route()
        app.openapi = self

    def register(self, app):
        """
        :param app: `flask.Flask`

        Register to Flask application.

        This function will update some config, and add some routes to Flask
        application. If you have already pass the Flask application when
        Flaskerk is initialized, then it will register automatically.
        """
        self._init_app(app)

    def update_config(self, **kwargs):
        """
        Manually update config.

        This function will be triggered when you register this library to Flask
        instance, and configs in Flask.config['OPENAPI'] will be used to update.
        """
        for key, value in kwargs.items():
            setattr(self.config, key, value)

    def _register_route(self):
        """
        register doc blueprint to Flask app
        """
        blueprint = Blueprint(
            self.config.name,
            __name__,
            url_prefix=self.config.url_prefix,
            template_folder=self.config.template_folder,
        )

        # docs
        blueprint.add_url_rule(
            self.config.endpoint,
            self.config.name,
            view_func=APIview().as_view(
                self.config.name,
                view_args=dict(config=self.config),
            )
        )

        # docs/openapi.json
        @blueprint.route(f'{self.config.endpoint}<filename>')
        def jsonfile(filename):
            if filename == self.config.filename:
                return jsonify(self.spec)
            abort(404)

        self.app.register_blueprint(blueprint)

    def bypass(self, func):
        if self.config.mode == 'greedy':
            return False
        elif self.config.mode == 'strict':
            if getattr(func, '_decorator', None) == self:
                return False
            return True
        else:
            decorator = getattr(func, '_decorator', None)
            if decorator and decorator != self:
                return True
            return False

    @property
    def spec(self):
        """
        Get OpenAPI spec for this Flask app.
        """
        if self.config._spec is None:
            self._generate_spec()
        return self.config._spec

    def _generate_spec(self):
        """
        generate OpenAPI spec JSON file
        """
        routes = {}
        for rule in self.app.url_map.iter_rules():
            if str(rule).startswith(self.config.endpoint) or \
                    str(rule).startswith('/static'):
                continue

            func = self.app.view_functions[rule.endpoint]
            path, parameters = parse_url(str(rule))

            # bypass the function decorated by others
            if self.bypass(func):
                continue

            # multiple methods (with different func) may bond to the same path
            if path not in routes:
                routes[path] = {}

            for method in rule.methods:
                if method in ['HEAD', 'OPTIONS']:
                    continue

                spec = {
                    'summary': func.__name__.capitalize(),
                    'operationID': func.__name__ + '__' + method.lower(),
                }

                if hasattr(func, 'data'):
                    spec['requestBody'] = {
                        'content': {
                            'application/json': {
                                'schema': {
                                    '$ref': f'#/components/schemas/{func.data}'
                                }
                            }
                        }
                    }

                if hasattr(func, 'query'):
                    parameters.append({
                        'name': func.query,
                        'in': 'query',
                        'required': True,
                        'schema': {
                            '$ref': f'#/components/schemas/{func.query}',
                        }
                    })
                spec['parameters'] = parameters

                spec['responses'] = {}
                has_2xx = False
                if hasattr(func, 'x'):
                    for code, msg in func.x.items():
                        if code.startswith('2'):
                            has_2xx = True
                        spec['responses'][code] = {
                            'description': msg,
                        }

                if hasattr(func, 'resp'):
                    spec['responses']['200'] = {
                        'description': 'Successful Response',
                        'content': {
                            'application/json': {
                                'schema': {
                                    '$ref': f'#/components/schemas/{func.resp}'
                                }
                            }
                        },
                    }
                elif not has_2xx:
                    spec['responses']['200'] = { 'description': 'Successful Response' }

                if any([hasattr(func, schema)
                        for schema in ('query', 'data', 'resp')]):
                    spec['responses']['422'] = {
                        'description': 'Validation Error',
                    }

                routes[path][method.lower()] = spec

        definitions = {}
        for _, schema in self.models.items():
            if 'definitions' in schema:
                for key, value in schema['definitions'].items():
                    definitions[key] = value
                del schema['definitions']

        data = {
            'openapi': self.config.openapi_veresion,
            'info': {
                'title': self.config.title,
                'version': self.config.version,
            },
            'paths': {
                **routes
            },
            'components': {
                'schemas': {
                    name: schema for name, schema in self.models.items()
                },
            },
            'definitions': definitions
        }
        self.config._spec = data

    def validate(self, query=None, data=None, resp=None, x=[]):
        """
        :param query: ``pydantic.BaseModel`` schema for request. The parsed
                      data will store in :class:`flask.request.query`.
        :param resp: ``pydantic.BaseModel`` schema for response
        :param data: ``pydantic.BaseModel`` schema for JSON data
        :param x: List of :class:`flaskerk.exception.HTTPException`

        validate JSON data according to Model schema

        .. code-block:: python

           from pydantic import BaseModel
           from flask import request, jsonify

           class Query(BaseModel):
               text: str
               limit: int

           @app.route('/api/predict', methods=['POST'])
           @api.validate(data=Query)
           def predict():
               data = request.json_data
               print(data.text, data.limit)
               return jsonify(is_spam=True)

        For more examples, check examples_.

        .. _examples: https://github.com/kemingy/flaskerk/tree/master/examples
        """
        def decorate_validate_request(func):
            @wraps(func)
            def validate_request(*args, **kwargs):
                try:
                    # validate query
                    arg = request.args
                    if not arg:
                        arg = {}
                    json_query = query(**arg) if query else None

                    # validate data
                    json_obj = request.get_json()
                    if json_obj is None:
                        json_obj = {}
                    json_data = data(**json_obj) if data else None
                except ValidationError as err:
                    abort(make_response(jsonify(message=str(err)), 422))
                except Exception:
                    raise

                request.query = json_query
                request.json_data = json_data

                response = func(*args, **kwargs)
                others = ()
                if isinstance(response, tuple) and len(response) > 1:
                    response, others = response[0], response[1:]
                if resp and not isinstance(response, resp):
                    abort_json(500, 'Wrong response type produced by server.')

                if resp:
                    return make_response(jsonify(**response.dict()), *others)
                return make_response(response, *others)

            # register schemas to this function
            for schema, name in zip(
                (query, data, resp), ('query', 'data', 'resp')
            ):
                if schema:
                    assert issubclass(schema, BaseModel)
                    self.models[schema.__name__] = schema.schema()
                    setattr(validate_request, name, schema.__name__)

            # store exception for doc
            code_msg = {}
            for e in x:
                assert isinstance(e, HTTPException)
                code_msg[str(e.code)] = e.msg

            if code_msg:
                validate_request.x = code_msg

            # register this decorator
            validate_request._decorator = self

            return validate_request
        return decorate_validate_request



from setuptools import setup, find_packages
from os import path
from io import open


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requires = [req.strip() for req in f if req]


setup(
    name='flaskerk',
    version='0.6.0',
    author='Keming Yang',
    author_email='kemingy94@gmail.com',
    description='',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/kemingy/flaskerk',
    packages=find_packages(exclude=['examples*', 'tests*']),
    package_data={
        'flaskerk': ['templates/*.html'],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    install_requires=requires,
    zip_safe=False,
    extras_require={},
    entry_points={
        'console_scripts': [],
    },
)



class Config:
    """
    :ivar endpoint: url path for docs
    :ivar filename: openapi spec file name
    :ivar openapi_version: openapi spec version
    :ivar title: document title
    :ivar version: service version
    :ivar ui: ui theme, choose 'redoc' or 'swagger'
    :ivar mode: mode for route. **normal** includes undecorated routes and
        routes decorated by this instance. **strict** only includes routes
        decorated by this instance. **greedy** includes all the routes.

    Flaskerk configuration.
    """

    def __init__(self):
        self.name = 'docs'
        self.endpoint = '/docs/'
        self.url_prefix = None
        self.template_folder = 'templates'
        self.filename = 'openapi.json'
        self.mode = 'normal'

        self.openapi_veresion = '3.0.2'
        self.title = 'Service Documents'
        self.version = 'latest'
        self.ui = 'redoc'

        self._support_ui = {'redoc', 'swagger'}
        self._support_mode = {'normal', 'greedy', 'strict'}
